import os
import cv2
import glob
import random
from PIL import Image
import numpy as np
import torch
from torchvision.io import read_video, write_jpeg
from torch.utils.data import Dataset
from torchvision import transforms as T
import imgaug.augmenters as iaa
import albumentations as A
from perlin import rand_perlin_2d_np
from nsa import patch_ex
from skimage.segmentation import mark_boundaries


class CustomDataset(Dataset):
    def __init__(self, c, is_train=True, excluded_images=None):
        self.dataset_path = c.data_path
        self.is_train = is_train
        self.cropsize = c.crop_size
        self.train_list = c.train_list
        self.val_list = c.val_list
        # load dataset
        if excluded_images is not None:
            self.x, self.y, self.mask, self.img_types = self.load_dataset_list()
            self.x, self.y, self.mask, self.img_types = excluding_images(self.x, self.y, self.mask, self.img_types, excluded_images)
        else:
            self.x, self.y, self.mask, self.img_types = self.load_dataset_list()
        # set transforms
        if is_train:
            self.transform_x = T.Compose([
                T.Resize(c.img_size, Image.ANTIALIAS),
                #T.RandomRotation(5),
                # T.CenterCrop(c.crop_size),
                T.ToTensor()])
        # test:
        else:
            self.transform_x = T.Compose([
                T.Resize(c.img_size, Image.ANTIALIAS),
                # T.CenterCrop(c.crop_size),
                T.ToTensor()])
        # mask
        self.transform_mask = T.Compose([
            T.Resize(c.img_size, Image.NEAREST),
            # T.CenterCrop(c.crop_size),
            T.ToTensor()])

        self.normalize = T.Compose([T.Normalize(c.norm_mean, c.norm_std)])

    def __getitem__(self, idx):
        img_path, y, mask, img_type = self.x[idx], self.y[idx], self.mask[idx], self.img_types[idx]
        
        x = Image.open(img_path)
        x = self.normalize(self.transform_x(x))
        
        if y == 0:
            mask = torch.zeros([1, self.cropsize[0], self.cropsize[1]])
        else:
            mask = Image.open(mask)
            mask = self.transform_mask(mask)
        return x, y, mask, os.path.basename(img_path[:-4]), img_type

    def __len__(self):
        return len(self.x)

    def load_dataset_list(self):
        x, y, mask, types = [], [], [], []
        x_n,y_n,mask_n,types_n = [],[],[],[]
        x_ab,y_ab,mask_ab,types_ab = [],[],[],[]
        
        # [img_path, normal(0) or NG(1), mask_path, type_name]
        if self.is_train == True:
            for t in self.train_list:
                if t[1] == 0:
                    x_n.extend([t[0]])
                    y_n.extend([t[1]])
                    mask_n.extend([t[2]])
                    types_n.extend([t[3]])
            x = x_n
            y = y_n
            mask = mask_n
            types = types_n
        elif self.is_train== False:
            for v in self.val_list:
                if v[1] == 0:
                    x_n.extend([v[0]])
                    y_n.extend([v[1]])
                    mask_n.extend([v[2]])
                    types_n.extend([v[3]])
                if v[1] == 1:
                    x_ab.extend([v[0]])
                    y_ab.extend([v[1]])
                    mask_ab.extend([v[2]])
                    types_ab.extend([v[3]])
            x = x_n + x_ab
            y = y_n + y_ab
            mask = mask_n + mask_ab
            types = types_n + types_ab
        
        return list(x), list(y), list(mask), list(types)
        
        
class CustomFSCopyPasteDataset(Dataset):
    """
    Mvtec train dataset with anomaly samples.
    Anomaly samples: real anomaly samples,
        anomaly samples generated by copy-pasting abnormal regions to normal samples
    """
    def __init__(self, c, is_train=True):
        self.dataset_path = c.data_path
        self.is_train = is_train
        self.cropsize = c.crop_size
        self.anomaly_nums = c.num_anomalies  # number of anomaly samples in each abnormal type
        self.repeat_num = 10  # repeat times for anomaly samples
        self.reuse_times = 5  # real anomaly reuse times
        self.in_fg_region = c.in_fg_region
        
        self.train_list = c.train_list
        # load dataset
        self.n_imgs, self.n_labels, self.n_masks, self.a_imgs, self.a_labels, self.a_masks = self.load_dataset_list()
        self.a_imgs = self.a_imgs * self.repeat_num
        self.a_labels = self.a_labels * self.repeat_num
        self.a_masks = self.a_masks * self.repeat_num
        
        self.labels = np.array(self.n_labels + self.a_labels)
        self.normal_idx = np.argwhere(self.labels == 0).flatten()
        self.anomaly_idx = np.argwhere(self.labels == 1).flatten()
        
        self.img_size = c.img_size
        
        # set transforms
        if is_train:
            self.transform_img = T.Compose([
                T.Resize(c.img_size, Image.ANTIALIAS),
                #T.RandomRotation(5),
                T.CenterCrop(c.crop_size),
                T.ToTensor()])
        # test:
        else:
            self.transform_img = T.Compose([
                T.Resize(c.img_size, Image.ANTIALIAS),
                T.CenterCrop(c.crop_size),
                T.ToTensor()])
        # mask
        self.transform_mask = T.Compose([
            T.Resize(c.img_size, Image.NEAREST),
            T.CenterCrop(c.crop_size),
            T.ToTensor()])
        
        self.augmentors = [A.RandomRotate90(),
                A.Flip(),
                A.Transpose(),
                A.OpticalDistortion(p=1.0, distort_limit=1.0),
                A.OneOf([
                    A.IAAAdditiveGaussianNoise(),
                    A.GaussNoise(),
                ], p=0.2),
                A.OneOf([
                    A.MotionBlur(p=.2),
                    A.MedianBlur(blur_limit=3, p=0.1),
                    A.Blur(blur_limit=3, p=0.1),
                ], p=0.2),
                A.ShiftScaleRotate(shift_limit=0.0625, scale_limit=0.2, rotate_limit=45, p=0.2),
                A.OneOf([
                    A.OpticalDistortion(p=0.3),
                    A.GridDistortion(p=.1),
                    A.IAAPiecewiseAffine(p=0.3),
                ], p=0.2),
                A.OneOf([
                    A.CLAHE(clip_limit=2),
                    A.IAASharpen(),
                    A.IAAEmboss(),
                    A.RandomBrightnessContrast(),            
                ], p=0.3),
                A.HueSaturationValue(p=0.3)]

        self.normalize = T.Compose([T.Normalize(c.norm_mean, c.norm_std)])
    
    def __len__(self):
        return len(self.n_imgs) + len(self.a_imgs)

    def __getitem__(self, idx):
        if idx >= len(self.n_imgs):  # anomaly samples
            idx_ = idx - len(self.n_imgs)
            img, label, mask = self.a_imgs[idx_], self.a_labels[idx_], self.a_masks[idx_]
            if idx >= len(self.n_imgs) + self.anomaly_nums * self.reuse_times:
                # generating anomaly sample by copy-pasting
                img, mask = self.copy_paste(img, mask, self.img_size)
                img, mask = Image.fromarray(img), Image.fromarray(mask)
                # img.save('aug_imgs/gen_img.jpg')
                # mask.save('aug_imgs/gen_mask.png')
                img = self.normalize(self.transform_img(img))
                mask = self.transform_mask(mask)

                return img, label, mask
        else:  # normal samples
            img, label, mask = self.n_imgs[idx], self.n_labels[idx], self.n_masks[idx]
        img = Image.open(img)
        #
        img = self.normalize(self.transform_img(img))
        #
        if label == 0:
            mask = torch.zeros([1, self.cropsize[0], self.cropsize[1]])
        else:
            mask = Image.open(mask)
            mask = self.transform_mask(mask)
        
        return img, label, mask
    
    def randAugmenter(self):
        aug_ind = np.random.choice(np.arange(len(self.augmentors)), 3, replace=False)
        aug = A.Compose([self.augmentors[aug_ind[0]],
                         self.augmentors[aug_ind[1]],
                         self.augmentors[aug_ind[2]]])
        # aug = A.Compose([self.augmentors[0], A.GridDistortion(p=1.0), self.augmentors[3], self.augmentors[1], self.augmentors[7]])
        return aug

    def copy_paste(self, img, mask, img_size):
        n_idx = np.random.randint(len(self.n_imgs))  # get a random normal sample
        aug = self.randAugmenter()

        image = cv2.imread(img)  # anomaly sample
        image = cv2.resize(image,img_size)
        # temp_img = Image.open(img)
        # temp_img.save("aug_imgs/ano_img.jpg")
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)  # (900, 900, 3)
        n_image = cv2.imread(self.n_imgs[n_idx])  # normal sample
        n_image = cv2.resize(n_image,img_size)
        n_image = cv2.cvtColor(n_image, cv2.COLOR_BGR2RGB)  # (900, 900, 3)
        # temp = Image.open(self.n_imgs[n_idx])
        # temp.save('aug_imgs/nor_img.jpg')
        img_height, img_width = n_image.shape[0], n_image.shape[1]

        mask = Image.open(mask)
        mask = mask.resize(img_size)
        mask = np.asarray(mask)  # (900, 900)
        
        
        # augmente the abnormal region
        augmentated = aug(image=image, mask=mask)
        aug_image, aug_mask = augmentated['image'], augmentated['mask']
        # temp_img = Image.fromarray(aug_image)
        # temp_img.save("aug_imgs/ano_aug_img.jpg")
        # crop_img = aug_image.copy()
        # crop_img[aug_mask == 0] = 0
        # crop_img = Image.fromarray(crop_img)
        # crop_img.save('aug_imgs/crop_img.jpg')
        # if self.in_fg_region:
        #     n_img_path = self.n_imgs[n_idx]
        #     img_file = n_img_path.split('/')[-1]
        #     fg_path = os.path.join(f'fg_mask/{self.class_name}', img_file)
        #     fg_mask = Image.open(fg_path)
        #     #fg_mask.save("aug_imgs/fg_mask.jpg")
        #     fg_mask = np.asarray(fg_mask)
            
        #     intersect_mask = np.logical_and(fg_mask == 255, aug_mask == 255)
        #     if (np.sum(intersect_mask) > int(2 / 3 * np.sum(aug_mask == 255))):
        #         # when most part of aug_mask is in the fg_mask region 
        #         # copy the augmentated anomaly area to the normal image
        #         n_image[aug_mask == 255, :] = aug_image[aug_mask == 255, :]
        #         return n_image, aug_mask
        #     else:
        #         contours, _ = cv2.findContours(aug_mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        #         center_xs, center_ys = [], []
        #         widths, heights = [], []
        #         for i in range(len(contours)):
        #             M = cv2.moments(contours[i])
        #             if M['m00'] == 0:  # error case
        #                 x_min, x_max = np.min(contours[i][:, :, 0]), np.max(contours[i][:, :, 0])
        #                 y_min, y_max = np.min(contours[i][:, :, 1]), np.max(contours[i][:, :, 1])
        #                 center_x = int((x_min + x_max) / 2)
        #                 center_y = int((y_min + y_max) / 2)
        #             else:
        #                 center_x = int(M["m10"] / M["m00"])
        #                 center_y = int(M["m01"] / M["m00"])
        #             center_xs.append(center_x)
        #             center_ys.append(center_y)
        #             x_min, x_max = np.min(contours[i][:, :, 0]), np.max(contours[i][:, :, 0])
        #             y_min, y_max = np.min(contours[i][:, :, 1]), np.max(contours[i][:, :, 1])
        #             width, height = x_max - x_min, y_max - y_min
        #             widths.append(width)
        #             heights.append(height)
        #         if len(widths) == 0 or len(heights) == 0:  # no contours
        #             n_image[aug_mask == 255, :] = aug_image[aug_mask == 255, :]
        #             return n_image, aug_mask
        #         else:
        #             max_width, max_height = np.max(widths), np.max(heights)
        #             center_mask = np.zeros((img_height, img_width), dtype=np.uint8)
        #             center_mask[int(max_height/2):img_height-int(max_height/2), int(max_width/2):img_width-int(max_width/2)] = 255
        #             fg_mask = np.logical_and(fg_mask == 255, center_mask == 255)

        #             x_coord = np.arange(0, img_width)
        #             y_coord = np.arange(0, img_height)
        #             xx, yy = np.meshgrid(x_coord, y_coord)
        #             # coordinates of fg region points
        #             xx_fg = xx[fg_mask]
        #             yy_fg = yy[fg_mask]
        #             xx_yy_fg = np.stack([xx_fg, yy_fg], axis=-1)  # (N, 2)
                    
        #             if xx_yy_fg.shape[0] == 0:  # no fg
        #                 n_image[aug_mask == 255, :] = aug_image[aug_mask == 255, :]
        #                 return n_image, aug_mask

        #             aug_mask_shifted = np.zeros((img_height, img_width), dtype=np.uint8)
        #             for i in range(len(contours)):
        #                 aug_mask_shifted_i = np.zeros((img_height, img_width), dtype=np.uint8)
        #                 new_aug_mask_i = np.zeros((img_height, img_width), dtype=np.uint8)
        #                 # random generate a point in the fg region
        #                 idx = np.random.choice(np.arange(xx_yy_fg.shape[0]), 1)
        #                 rand_xy = xx_yy_fg[idx]
        #                 delta_x, delta_y = center_xs[i] - rand_xy[0, 0], center_ys[i] - rand_xy[0, 1]
                        
        #                 x_min, x_max = np.min(contours[i][:, :, 0]), np.max(contours[i][:, :, 0])
        #                 y_min, y_max = np.min(contours[i][:, :, 1]), np.max(contours[i][:, :, 1])
                        
        #                 # mask for one anomaly region
        #                 aug_mask_i = np.zeros((img_height, img_width), dtype=np.uint8)
        #                 aug_mask_i[y_min:y_max, x_min:x_max] = 255
        #                 aug_mask_i = np.logical_and(aug_mask == 255, aug_mask_i == 255)
                        
        #                 # coordinates of orginal mask points
        #                 xx_ano, yy_ano = xx[aug_mask_i], yy[aug_mask_i]
                        
        #                 # shift the original mask into fg region
        #                 xx_ano_shifted = xx_ano - delta_x
        #                 yy_ano_shifted = yy_ano - delta_y
        #                 outer_points_x = np.logical_or(xx_ano_shifted < 0, xx_ano_shifted >= img_width) 
        #                 outer_points_y = np.logical_or(yy_ano_shifted < 0, yy_ano_shifted >= img_height)
        #                 outer_points = np.logical_or(outer_points_x, outer_points_y)
                        
        #                 # keep points in image
        #                 xx_ano_shifted = xx_ano_shifted[~outer_points]
        #                 yy_ano_shifted = yy_ano_shifted[~outer_points]
        #                 aug_mask_shifted_i[yy_ano_shifted, xx_ano_shifted] = 255
                        
        #                 # original points should be changed
        #                 xx_ano = xx_ano[~outer_points]
        #                 yy_ano = yy_ano[~outer_points]
        #                 new_aug_mask_i[yy_ano, xx_ano] = 255
        #                 # copy the augmentated anomaly area to the normal image
        #                 n_image[aug_mask_shifted_i == 255, :] = aug_image[new_aug_mask_i == 255, :]
        #                 aug_mask_shifted[aug_mask_shifted_i == 255] = 255
        #             return n_image, aug_mask_shifted
        # else:  # no fg restriction
        #     # copy the augmentated anomaly area to the normal image
        n_image[aug_mask == 255, :] = aug_image[aug_mask == 255, :]

        return n_image, aug_mask

    def load_dataset_list(self):
        n_img_paths, n_labels, n_mask_paths = [], [], []  # normal
        a_img_paths, a_labels, a_mask_paths = [], [], []  # abnormal

        for t in self.train_list:
            if t[1] == 0:
                n_img_paths.extend([t[0]])
                n_labels.extend([t[1]])
                n_mask_paths.extend([t[2]])
            if t[1] == 1:
                a_img_paths.extend([t[0]])
                a_labels.extend([t[1]])
                a_mask_paths.extend([t[2]])
        
        
        return n_img_paths, n_labels, n_mask_paths, a_img_paths, a_labels, a_mask_paths



def excluding_images(images, labels, masks, img_types, excluding_images):
    retain_images = []
    retain_labels = []
    retain_masks = []
    retain_img_types = []
    for idx, image in enumerate(images):
        if image in excluding_images:
            continue
        retain_images.append(image)
        retain_labels.append(labels[idx])
        retain_masks.append(masks[idx])
        retain_img_types.append(img_types[idx])
    
    return retain_images, retain_labels, retain_masks, retain_img_types


from fc_flow import flow_model, conditional_flow_model
from modules import positionalencoding2d


def load_flow_model(args, in_channels):
    if args.flow_arch == 'flow_model':
        model = flow_model(args, in_channels)
    elif args.flow_arch == 'conditional_flow_model':
        model = conditional_flow_model(args, in_channels)
    else:
        raise NotImplementedError('{} is not supported Normalizing Flow!'.format(args.flow_arch))
    
    return model


import numpy as np
from torch.utils.data import Sampler


def excluding_images(images, labels, masks, img_types, excluding_images):
    retain_images = []
    retain_labels = []
    retain_masks = []
    retain_img_types = []
    for idx, image in enumerate(images):
        if image in excluding_images:
            continue
        retain_images.append(image)
        retain_labels.append(labels[idx])
        retain_masks.append(masks[idx])
        retain_img_types.append(img_types[idx])
    
    return retain_images, retain_labels, retain_masks, retain_img_types


class BalancedBatchSampler(Sampler):
    def __init__(self,
                 cfg,
                 dataset):
        super(BalancedBatchSampler, self).__init__(dataset)
        self.cfg = cfg
        self.dataset = dataset

        self.normal_generator = self.randomGenerator(self.dataset.normal_idx)
        self.outlier_generator = self.randomGenerator(self.dataset.anomaly_idx)
        # n_normal: 2/3; n_outlier: 1/3
        if self.cfg.num_anomalies != 0:
            self.n_normal = 2 * self.cfg.batch_size // 3
            self.n_anomaly = self.cfg.batch_size - self.n_normal
        else:
            self.n_normal = self.cfg.batch_size
            self.n_anomaly = 0

    def randomGenerator(self, list):
        while True:
            random_list = np.random.permutation(list)
            for i in random_list:
                yield i
    
    def __len__(self):
        return self.cfg.steps_per_epoch
    
    def __iter__(self):
        for _ in range(self.cfg.steps_per_epoch):
            batch = []

            for _ in range(self.n_normal):
                batch.append(next(self.normal_generator))

            for _ in range(self.n_anomaly):
                batch.append(next(self.outlier_generator))

            yield batch


import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np


def normal_fl_weighting(logps, gamma=0.5, alpha=11.7, normalizer=10):
    """
    Normal focal weighting.
    Args:
        logps: og-likelihoods, shape (N, ).
        gamma: gamma hyperparameter for normal focal weighting.
        alpha: alpha hyperparameter for abnormal focal weighting.
    """
    logps = logps / normalizer
    mask_larger = logps > -0.2
    mask_lower = logps <= -0.2
    probs = torch.exp(logps)
    fl_weights = alpha * (1 - probs).pow(gamma) * torch.abs(logps)
    weights = fl_weights.new_zeros(fl_weights.shape)
    weights[mask_larger] = 1.0 
    weights[mask_lower] = fl_weights[mask_lower]

    return weights


def abnormal_fl_weighting(logps, gamma=2, alpha=0.53, normalizer=10):
    """
    Abnormal focal weighting.
    Args:
        logps: og-likelihoods, shape (N, ).
        gamma: gamma hyperparameter for normal focal weighting.
        alpha: alpha hyperparameter for abnormal focal weighting.
    """
    logps = logps / normalizer
    mask_larger = logps > -1.0
    mask_lower = logps <= -1.0
    probs = torch.exp(logps)
    fl_weights = alpha * (1 + probs).pow(gamma) * (1 / torch.abs(logps))
    weights = fl_weights.new_zeros(fl_weights.shape)
    weights[mask_lower] = 1.0 
    weights[mask_larger] = fl_weights[mask_larger]

    return weights


def get_logp_boundary(logps, mask, pos_beta=0.05, margin_tau=0.1, normalizer=10):
    """
    Find the equivalent log-likelihood decision boundaries from normal log-likelihood distribution.
    Args:
        logps: log-likelihoods, shape (N, )
        mask: 0 for normal, 1 for abnormal, shape (N, )
        pos_beta: position hyperparameter: beta
        margin_tau: margin hyperparameter: tau
    """
    normal_logps = logps[mask == 0].detach()
    n_idx = int(((mask == 0).sum() * pos_beta).item())
    sorted_indices = torch.sort(normal_logps)[1]
    
    n_idx = sorted_indices[n_idx]
    b_n = normal_logps[n_idx]  # normal boundary
    b_n = b_n / normalizer

    b_a = b_n - margin_tau  # abnormal boundary

    return b_n, b_a


def calculate_bg_spp_loss(logps, mask, boundaries, normalizer=10, weights=None):
    """
    Calculate boudary guided semi-push-pull contrastive loss.
    Args:
        logps: log-likelihoods, shape (N, )
        mask: 0 for normal, 1 for abnormal, shape (N, 1)
        boundaries: normal boundary and abnormal boundary
    """
    logps = logps / normalizer
    b_n = boundaries[0]  # normal boundaries
    normal_logps = logps[mask == 0]
    normal_logps_inter = normal_logps[normal_logps <= b_n]
    loss_n = b_n - normal_logps_inter

    b_a = boundaries[1]
    anomaly_logps = logps[mask == 1]    
    anomaly_logps_inter = anomaly_logps[anomaly_logps >= b_a]
    loss_a = anomaly_logps_inter - b_a

    if weights is not None:
        nor_weights = weights[mask == 0][normal_logps <= b_n]
        loss_n = loss_n * nor_weights
        ano_weights = weights[mask == 1][anomaly_logps >= b_a]
        loss_a = loss_a * ano_weights
    
    loss_n = torch.mean(loss_n)
    loss_a = torch.mean(loss_a)

    return loss_n, loss_a


def calculate_bg_spp_loss_normal(logps, mask, boundaries, normalizer=10, weights=None):
    """
    Calculate boudary guided semi-push-pull contrastive loss.
    Args:
        logps: log-likelihoods, shape (N, )
        mask: 0 for normal, 1 for abnormal, shape (N, 1)
        boundaries: normal boundary and abnormal boundary
    """
    logps = logps / normalizer
    b_n = boundaries[0]  # normal boundaries
    normal_logps = logps[mask == 0]
    normal_logps_inter = normal_logps[normal_logps <= b_n]
    loss_n = b_n - normal_logps_inter

    if weights is not None:
        nor_weights = weights[mask == 0][normal_logps <= b_n]
        loss_n = loss_n * nor_weights
    
    loss_n = torch.mean(loss_n)

    return loss_n


# __init__.py
# Copyright (C) 2019 (gnyontu39@gmail.com) and contributors
#

import inspect
import os
import sys

__version__ = '0.6.6.1'

real_path = os.path.dirname(os.path.abspath(__file__)).replace("\\","/")
sys.path.append(real_path)

import bgad_fas_train_engine
import bgad_test_engine
import config
import custom_list
import datasets_init
import fc_flow
import losses
import main_utils
import model_utils
import models_init
import modules
import nsa
import perlin
import process_log
import train
import testing
import utils_init
import utils
import visualizer
import inferencing

__all__ = [name for name, obj in locals().items()
           if not (name.startswith('_') or inspect.ismodule(obj))]



import os
import math
import timm
import torch
import numpy as np
import torch.nn.functional as F
from sklearn.metrics import roc_auc_score
from utils_init import t2np, get_logp, load_weights
from datasets_init import create_test_data_loader
from models_init import positionalencoding2d, load_flow_model
from visualizer import plot_visualizing_results
from utils import calculate_pro_metric, convert_to_anomaly_scores, evaluate_thresholds,convert_to_anomaly_inference_scores
from process_log import test_state
from config import parse_args
from utils import init_seeds
from PIL import Image
import torchvision.transforms as transforms


def validate(args, image_path, encoder, decoders,max_values):
    
    decoders = [decoder.eval() for decoder in decoders]
    
    image_list, gt_label_list, gt_mask_list, file_names, img_types = [], [], [], [], []
    logps_list = [list() for _ in range(args.feature_levels)]
    with torch.no_grad():
        
        img = Image.open(image_path)
        
        preprocess = transforms.Compose([
            transforms.Resize(args.img_size,Image.ANTIALIAS),
            transforms.ToTensor(),
        ])
        normalize = transforms.Compose([transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])])
        
        input_tensor = normalize(preprocess(img))
        img = input_tensor.unsqueeze(0)
        
        image = img.to(args.device) # single scale
        features = encoder(image)  # BxCxHxW
        for l in range(args.feature_levels):
            e = features[l]  # BxCxHxW
            bs, dim, h, w = e.size()
            e = e.permute(0, 2, 3, 1).reshape(-1, dim)
            
            # (bs, 128, h, w)
            pos_embed = positionalencoding2d(args.pos_embed_dim, h, w).to(args.device).unsqueeze(0).repeat(bs, 1, 1, 1)
            pos_embed = pos_embed.permute(0, 2, 3, 1).reshape(-1, args.pos_embed_dim)
            decoder = decoders[l]

            if args.flow_arch == 'flow_model':
                z, log_jac_det = decoder(e)  
            else:
                z, log_jac_det = decoder(e, [pos_embed, ])

            logps = get_logp(dim, z, log_jac_det)
            logps = logps / dim  
            logps_list[l].append(logps.reshape(bs, h, w))
    
    scores = convert_to_anomaly_inference_scores(args, logps_list,max_values)
    # calculate detection AUROC
    img_scores = np.max(scores)
    if args.vis:
        # img_threshold, pix_threshold = evaluate_thresholds(gt_label, gt_mask, img_scores, scores)
        # save_dir = os.path.join(args.output_dir, args.exp_name, 'vis_results', args.class_name)
        # save_dir = os.path.join('vis_results', args.class_name)
        save_dir = args.heatmap_path
        os.makedirs(save_dir, exist_ok=True)
        plot_visualizing_results(image_list, scores, save_dir, file_names, args)
    
    return img_scores


def start_inference(device,gpu,image_path,encoder,decoders,input_size,save_heatmap,heatmap_threshold,save_heatmap_path,max_values):
    args = parse_args()
    if device == "GPU":
        args.device = torch.device("cuda")
    args.img_size = (input_size,input_size)
    args.crop_size = (input_size, input_size) 
    args.norm_mean, args.norm_std = [0.485, 0.456, 0.406], [0.229, 0.224, 0.225]
    args.img_dims = [3] + list(args.img_size)
    
    # data loaders
    # test_loader = create_test_data_loader(args)
    
    
    img_scores = validate(args, image_path, encoder, decoders,max_values)
    
    return img_scores

def load_model(device,gpu,checkpoint):
    init_seeds()
    args = parse_args()
    args.checkpoint = checkpoint
    
    os.environ['CUDA_VISIBLE_DEVICES'] = gpu
    if device == "GPU":
        args.device = torch.device("cuda")
        
    encoder = timm.create_model(args.backbone_arch, features_only=True, 
                out_indices=[i+1 for i in range(args.feature_levels)], pretrained=True)
    encoder = encoder.to(args.device).eval()
    feat_dims = encoder.feature_info.channels()
    
    # Normalizing Flows
    decoders = [load_flow_model(args, feat_dim) for feat_dim in feat_dims]
    decoders = [decoder.to(args.device) for decoder in decoders]
    
    load_weights(encoder, decoders, args.checkpoint)
    
    return encoder,decoders


import os
import math
import timm
import torch
import numpy as np
import torch.nn.functional as F
from sklearn.metrics import roc_auc_score
from utils_init import t2np, get_logp, load_weights
from datasets_init import create_test_data_loader
from models_init import positionalencoding2d, load_flow_model
from visualizer import plot_visualizing_results
from utils import calculate_pro_metric, convert_to_anomaly_scores, evaluate_thresholds
from process_log import test_state

def validate(args, data_loader, encoder, decoders):
    
    decoders = [decoder.eval() for decoder in decoders]
    
    image_list, gt_label_list, gt_mask_list, file_names, img_types = [], [], [], [], []
    logps_list = [list() for _ in range(args.feature_levels)]
    with torch.no_grad():
        for i, (image, label, mask, file_name, img_type) in enumerate(data_loader):
            test_state(i,len(data_loader))

            image_list.extend(t2np(image))
            file_names.extend(file_name)
            img_types.extend(img_type)
            gt_label_list.extend(t2np(label))
            gt_mask_list.extend(t2np(mask))
            
            image = image.to(args.device) # single scale
            features = encoder(image)  # BxCxHxW
            
            for l in range(args.feature_levels):
                e = features[l]  # BxCxHxW
                bs, dim, h, w = e.size()
                e = e.permute(0, 2, 3, 1).reshape(-1, dim)
               
                # (bs, 128, h, w)
                pos_embed = positionalencoding2d(args.pos_embed_dim, h, w).to(args.device).unsqueeze(0).repeat(bs, 1, 1, 1)
                pos_embed = pos_embed.permute(0, 2, 3, 1).reshape(-1, args.pos_embed_dim)
                decoder = decoders[l]

                if args.flow_arch == 'flow_model':
                    z, log_jac_det = decoder(e)  
                else:
                    z, log_jac_det = decoder(e, [pos_embed, ])

                logps = get_logp(dim, z, log_jac_det)  
                logps = logps / dim  
                logps_list[l].append(logps.reshape(bs, h, w))
    
    scores, max_values = convert_to_anomaly_scores(args, logps_list)
    # calculate detection AUROC
    img_scores = np.max(scores, axis=(1, 2))
    gt_label = np.asarray(gt_label_list, dtype=bool)
    img_auc = roc_auc_score(gt_label, img_scores)
    # calculate segmentation AUROC
    gt_mask = np.squeeze(np.asarray(gt_mask_list, dtype=bool), axis=1)
    pix_auc = roc_auc_score(gt_mask.flatten(), scores.flatten())
    if args.vis:
        img_threshold, pix_threshold = evaluate_thresholds(gt_label, gt_mask, img_scores, scores)
        # save_dir = os.path.join(args.output_dir, args.exp_name, 'vis_results', args.class_name)
        # save_dir = os.path.join('vis_results', args.class_name)
        save_dir = args.heatmap_path
        print(img_threshold)
        # os.makedirs(save_dir, exist_ok=True)
        plot_visualizing_results(image_list, scores, img_scores, gt_mask_list, pix_threshold, 
                                 img_threshold, save_dir, file_names, img_types,args)

    
    return img_auc, pix_auc, img_scores, file_names, max_values


def test(args):
    # Feature Extractor
    encoder = timm.create_model(args.backbone_arch, features_only=True, 
                out_indices=[i+1 for i in range(args.feature_levels)], pretrained=True)
    encoder = encoder.to(args.device).eval()
    feat_dims = encoder.feature_info.channels()
    
    # Normalizing Flows
    decoders = [load_flow_model(args, feat_dim) for feat_dim in feat_dims]
    decoders = [decoder.to(args.device) for decoder in decoders]

    # data loaders
    test_loader = create_test_data_loader(args)

    # Load weights(checkpoint)
    load_weights(encoder, decoders, args.checkpoint)

    img_auc, pix_auc, img_scores, file_names, max_values = validate(args, test_loader, encoder, decoders)
    origin_scores = []
    for v in args.val_list:
        origin_scores.append(img_scores[file_names.index(os.path.basename(v[0][:-4]))])
    origin_scores = np.array(origin_scores)
    return img_auc, pix_auc, origin_scores, max_values


import os
import numpy as np
import torch
import warnings
from config import parse_args
from utils import init_seeds
from PIL import Image

def main_single(device,gpu,input_size,val_list,checkpoint,save_heatmap,save_heatmap_path,heatmap_threshold):
    init_seeds()
    args = parse_args()

    # setting cuda 
    os.environ['CUDA_VISIBLE_DEVICES'] = gpu
    if device == "GPU":
        args.device = torch.device("cuda")
    # model path
    args.model_path = "{}_{}_{}_{}".format(
        args.dataset, args.backbone_arch, args.flow_arch, args.class_name)
    
    # image
    val_sample = Image.open(val_list[0][0])
    args.origin_size = val_sample.size
    # args.img_size = (args.inp_size, args.inp_size)  
    args.img_size = (input_size, input_size)  
    args.crop_size = (input_size, input_size) 
    args.norm_mean, args.norm_std = [0.485, 0.456, 0.406], [0.229, 0.224, 0.225]
    
    args.img_dims = [3] + list(args.img_size)

    # output settings
    args.save_results = True
    args.checkpoint = checkpoint
    args.heatmap_path = save_heatmap_path
    ############################################################################################################################################################
    # set data
    args.train_list = None
    args.val_list = val_list
    args.phase = "test"
    args.vis = save_heatmap
    
    args.heatmap_threshold = heatmap_threshold
    ############################################################################################################################################################
    from bgad_test_engine import test
    img_auc, pix_auc, img_scores, max_values = test(args)
    return img_auc, pix_auc, img_scores, max_values



from model_utils import save_results, save_weights, load_weights, adjust_learning_rate, warmup_learning_rate
from visualizer import denormalization, export_groundtruth, export_hist, export_scores, export_test_images, plot_visualizing_results
from utils import MetricRecorder, EachEpochRecorder, get_logp, t2np, rescale, calculate_pro_metric, evaluate_thresholds, convert_to_anomaly_scores


__all__ = ['save_results',
           'save_weights',
           'load_weights',
           'adjust_learning_rate',
           'warmup_learning_rate',
           'denormalization',
           'export_groundtruth',
           'export_hist',
           'export_scores',
           'export_test_images',
           'plot_visualizing_results',
           'MetricRecorder',
           'EachEpochRecorder',
           'get_logp',
           't2np',
           'rescale',
           'calculate_pro_metric',
           'evaluate_thresholds',
           'convert_to_anomaly_scores']


import numpy as np
import cv2
import sys
from skimage.morphology import disk
from skimage.filters import median


def patch_ex(ima_dest, ima_src=None, same=False, num_patches=1,
             mode=cv2.NORMAL_CLONE, width_bounds_pct=((0.05,0.2),(0.05,0.2)), min_object_pct=0.25, 
             min_overlap_pct=0.25, shift=True, label_mode='binary', skip_background=None, tol=1, resize=True,
             gamma_params=None, intensity_logistic_params=(1/6, 20),
             resize_bounds=(0.7, 1.3), verbose=True):
    """
    Create a synthetic training example from the given images by pasting/blending random patches.
    Args:
        ima_dest (uint8 numpy array): image with shape (W,H,3) or (W,H,1) where patch should be changed
        ima_src (uint8 numpy array): optional, otherwise use ima_dest as source
        same (bool): use ima_dest as source even if ima_src given
        mode: 'uniform', 'swap', 'mix', cv2.NORMAL_CLONE, or cv2.MIXED_CLONE what blending method to use
             ('mix' is flip a coin between normal and mixed clone)
        num_patches (int): how many patches to add. the method will always attempt to add the first patch,
                    for each subsequent patch it flips a coin
        width_bounds_pct ((float, float), (float, float)): min half-width of patch ((min_dim1, max_dim1), (min_dim2, max_dim2))
        shift (bool): if false, patches in src and dest image have same coords. otherwise random shift
        resize (bool): if true, patch is resampled at random size (within bounds and keeping aspect ratio the same) before blending  
        skip_background (int, int) or [(int, int),]: optional, assume background color is first and only interpolate patches
                    in areas where dest or src patch has pixelwise MAD < second from background.
        tol (int): mean abs intensity change required to get positive label
        gamma_params (float, float, float): optional, (shape, scale, left offset) of gamma dist to sample half-width of patch from,
                    otherwise use uniform dist between 0.05 and 0.95
        intensity_logistic_params (float, float): k, x0 of logitistc map for intensity based label
        label_mode: 'binary', 
                    'continuous' -- use interpolation factor as label (only when mode is 'uniform'),
                    'intensity' -- use median filtered mean absolute pixelwise intensity difference as label,
                    'logistic-intensity' -- use logistic median filtered of mean absolute pixelwise intensity difference as label,
    """
    if mode == 'mix':
        mode = (cv2.NORMAL_CLONE, cv2.MIXED_CLONE)[np.random.randint(2)]

    ima_src = ima_dest.copy() if same or (ima_src is None) else ima_src

    if skip_background is not None:
        if isinstance(skip_background, tuple):
            skip_background = [skip_background]
        src_object_mask = np.ones_like(ima_src[...,0:1])
        dest_object_mask = np.ones_like(ima_dest[...,0:1])
        for background, threshold in skip_background:
            src_object_mask &= np.uint8(np.abs(ima_src.mean(axis=-1, keepdims=True) - background) > threshold)
            dest_object_mask &= np.uint8(np.abs(ima_dest.mean(axis=-1, keepdims=True) - background) > threshold)
        src_object_mask[...,0] = cv2.medianBlur(src_object_mask[...,0], 7)  # remove grain from threshold choice
        dest_object_mask[...,0] = cv2.medianBlur(dest_object_mask[...,0], 7)  # remove grain from threshold choice
    else:
        src_object_mask = None
        dest_object_mask = None
    
    # add patches
    mask = np.zeros_like(ima_dest[..., 0:1])  # single channel
    patchex = ima_dest.copy()
    coor_min_dim1, coor_max_dim1, coor_min_dim2, coor_max_dim2 = mask.shape[0] - 1, 0, mask.shape[1] - 1, 0 
    if label_mode == 'continuous':
        factor = np.random.uniform(0.05, 0.95)
    else:
        factor = 1
    for i in range(num_patches):
        if i == 0 or np.random.randint(2) > 0:  # at least one patch
            patchex, ((_coor_min_dim1, _coor_max_dim1), (_coor_min_dim2, _coor_max_dim2)), patch_mask = _patch_ex(
                patchex, ima_src, dest_object_mask, src_object_mask, mode, label_mode, shift, resize, width_bounds_pct, 
                gamma_params, min_object_pct, min_overlap_pct, factor, resize_bounds, verbose)
            if patch_mask is not None:
                mask[_coor_min_dim1:_coor_max_dim1,_coor_min_dim2:_coor_max_dim2] = patch_mask
                coor_min_dim1 = min(coor_min_dim1, _coor_min_dim1) 
                coor_max_dim1 = max(coor_max_dim1, _coor_max_dim1) 
                coor_min_dim2 = min(coor_min_dim2, _coor_min_dim2)
                coor_max_dim2 = max(coor_max_dim2, _coor_max_dim2)

    # create label
    label_mask = np.uint8(np.mean(np.abs(1.0 * mask*ima_dest - 1.0 * mask*patchex), axis=-1, keepdims=True) > tol)
    label_mask[...,0] = cv2.medianBlur(label_mask[...,0], 5)  # remove grain from threshold choice

    if label_mode == 'continuous':
        label = label_mask * factor
    elif label_mode in ['logistic-intensity', 'intensity']:
        k, x0 = intensity_logistic_params
        label = np.mean(np.abs(label_mask * ima_dest * 1.0 - label_mask * patchex * 1.0), axis=-1, keepdims=True)
        label[...,0] = median(label[...,0], disk(5))
        if label_mode == 'logistic-intensity':
            label = label_mask / (1 + np.exp(-k * (label - x0)))
    elif label_mode == 'binary':
        label = label_mask  
    else:
        raise ValueError("label_mode not supported" + str(label_mode))

    return patchex, label


def _patch_ex(ima_dest, ima_src, dest_object_mask, src_object_mask, mode, label_mode, shift, resize, width_bounds_pct, 
              gamma_params, min_object_pct, min_overlap_pct, factor, resize_bounds, verbose):
    skip_background = (src_object_mask is not None) and (dest_object_mask is not None)
    dims = np.array(ima_dest.shape)
    min_width_dim1 = (width_bounds_pct[0][0]*dims[0]).round().astype(int)
    max_width_dim1 = (width_bounds_pct[0][1]*dims[0]).round().astype(int)
    min_width_dim2 = (width_bounds_pct[1][0]*dims[1]).round().astype(int)
    max_width_dim2 = (width_bounds_pct[1][1]*dims[1]).round().astype(int)

    if gamma_params is not None:
        shape, scale, lower_bound = gamma_params
        patch_width_dim1 = int(np.clip((lower_bound + np.random.gamma(shape, scale)) * dims[0], min_width_dim1, max_width_dim1))
        patch_width_dim2 = int(np.clip((lower_bound + np.random.gamma(shape, scale)) * dims[1], min_width_dim2, max_width_dim2))
    else:
        patch_width_dim1 = np.random.randint(min_width_dim1, max_width_dim1)
        patch_width_dim2 = np.random.randint(min_width_dim2, max_width_dim2)

    found_patch = False
    attempts = 0
    while not found_patch:
        center_dim1 = np.random.randint(min_width_dim1, dims[0]-min_width_dim1)
        center_dim2 = np.random.randint(min_width_dim2, dims[1]-min_width_dim2)

        coor_min_dim1 = np.clip(center_dim1 - patch_width_dim1, 0, dims[0])
        coor_min_dim2 = np.clip(center_dim2 - patch_width_dim2, 0, dims[1])
        coor_max_dim1 = np.clip(center_dim1 + patch_width_dim1, 0, dims[0])
        coor_max_dim2 = np.clip(center_dim2 + patch_width_dim2, 0, dims[1])

        patch_mask = np.ones((coor_max_dim1 - coor_min_dim1, coor_max_dim2 - coor_min_dim2, 1), dtype=np.uint8) 

        if skip_background:
            background_area = np.sum(patch_mask & src_object_mask[coor_min_dim1:coor_max_dim1, coor_min_dim2:coor_max_dim2])
            found_patch = (background_area / (patch_mask.shape[0] * patch_mask.shape[1]) > min_object_pct) 
        else:
            found_patch = True
        attempts += 1
        if attempts == 200:
            if verbose:
                print('No suitable patch found.')
            return ima_dest.copy(), ((0,0),(0,0)), None

    src = ima_src[coor_min_dim1:coor_max_dim1, coor_min_dim2:coor_max_dim2]
    height, width, _ = src.shape
    if resize:
        lb, ub = resize_bounds
        scale = np.clip(np.random.normal(1, 0.5), lb, ub)
        new_height = np.clip(scale * height, min_width_dim1, max_width_dim1)
        new_width = np.clip(int(new_height / height * width), min_width_dim2, max_width_dim2)
        new_height = np.clip(int(new_width / width * height), min_width_dim1, max_width_dim1)  # in case there was clipping
        if src.shape[2] == 1:  # grayscale
            src = cv2.resize(src[..., 0], (new_width, new_height))
            src = src[...,None]
        else:
            src = cv2.resize(src, (new_width, new_height))
        height, width, _ = src.shape
        patch_mask = cv2.resize(patch_mask[...,0], (width, height))
        patch_mask = patch_mask[...,None]

    if skip_background:
        src_object_mask = cv2.resize(src_object_mask[coor_min_dim1:coor_max_dim1, coor_min_dim2:coor_max_dim2, 0], (width, height))
        src_object_mask = src_object_mask[...,None]
    
    # sample destination location and size
    if shift:
        found_center = False
        attempts = 0
        while not found_center:
            center_dim1 = np.random.randint(height//2 + 1, ima_dest.shape[0] - height//2 - 1)
            center_dim2 = np.random.randint(width//2 + 1, ima_dest.shape[1] - width//2 - 1)
            coor_min_dim1, coor_max_dim1 = center_dim1 - height//2, center_dim1 + (height+1)//2
            coor_min_dim2, coor_max_dim2 = center_dim2 - width//2, center_dim2 + (width+1)//2

            if skip_background: 
                src_and_dest = dest_object_mask[coor_min_dim1:coor_max_dim1, coor_min_dim2:coor_max_dim2] & src_object_mask & patch_mask
                src_or_dest = (dest_object_mask[coor_min_dim1:coor_max_dim1, coor_min_dim2:coor_max_dim2] | src_object_mask) & patch_mask
                found_center = (np.sum(src_object_mask) / (patch_mask.shape[0] * patch_mask.shape[1]) > min_object_pct and    # contains object
                            np.sum(src_and_dest) / np.sum(src_object_mask) > min_overlap_pct)                    # object overlaps src object
            else:
                found_center = True
            attempts += 1
            if attempts == 200:
                if verbose:
                    print('No suitable center found. Dims were:', width, height)
                return ima_dest.copy(), ((0,0),(0,0)), None
            
    # blend
    if skip_background:
        patch_mask &= src_object_mask | dest_object_mask[coor_min_dim1:coor_max_dim1, coor_min_dim2:coor_max_dim2] 

    if mode == 'swap':
        patchex = ima_dest.copy()
        before = patchex[coor_min_dim1:coor_max_dim1, coor_min_dim2:coor_max_dim2]
        patchex[coor_min_dim1:coor_max_dim1, coor_min_dim2:coor_max_dim2] -= patch_mask * before
        patchex[coor_min_dim1:coor_max_dim1, coor_min_dim2:coor_max_dim2] += patch_mask * src
    elif mode == 'uniform':
        patchex = 1.0 * ima_dest
        before = patchex[coor_min_dim1:coor_max_dim1, coor_min_dim2:coor_max_dim2]
        patchex[coor_min_dim1:coor_max_dim1, coor_min_dim2:coor_max_dim2] -= factor * patch_mask * before
        patchex[coor_min_dim1:coor_max_dim1, coor_min_dim2:coor_max_dim2] += factor * patch_mask * src
        patchex = np.uint8(np.floor(patchex))
    elif mode in [cv2.NORMAL_CLONE, cv2.MIXED_CLONE]:  # poisson interpolation
        int_factor = np.uint8(np.ceil(factor * 255))
        # add background to patchmask to avoid artefacts
        if skip_background:
            patch_mask_scaled = int_factor * (patch_mask | ((1 - src_object_mask) & (1 - dest_object_mask[coor_min_dim1:coor_max_dim1, coor_min_dim2:coor_max_dim2])))
        else:
            patch_mask_scaled = int_factor * patch_mask
        patch_mask_scaled[0], patch_mask_scaled[-1], patch_mask_scaled[:,0], patch_mask_scaled[:,-1] = 0, 0, 0, 0  # zero border to avoid artefacts
        center = (coor_max_dim2 - (coor_max_dim2 - coor_min_dim2) // 2, coor_min_dim1 + (coor_max_dim1 - coor_min_dim1) // 2)  # height dim first
        if np.sum(patch_mask_scaled > 0) < 50:  # cv2 seamlessClone will fail if positive mask area is too small
            return ima_dest.copy(), ((0,0),(0,0)), None
        try:
            if ima_dest.shape[2] == 1:  # grayscale
                # pad to 3 channels as that's what OpenCV expects
                src_3 = np.concatenate((src, np.zeros_like(src), np.zeros_like(src)), axis=2)
                ima_dest_3 = np.concatenate((ima_dest, np.zeros_like(ima_dest), np.zeros_like(ima_dest)), axis=2)
                patchex = cv2.seamlessClone(src_3, ima_dest_3, patch_mask_scaled, center, mode)
                patchex = patchex[...,0:1]  # extract first channel
            else:  # RGB
                patchex = cv2.seamlessClone(src, ima_dest, patch_mask_scaled, center, mode)
        except cv2.error as e:
            print('WARNING, tried bad interpolation mask and got:', e)
            return ima_dest.copy(), ((0,0),(0,0)), None
    else:
        raise ValueError("mode not supported" + str(mode))

    return patchex, ((coor_min_dim1, coor_max_dim1), (coor_min_dim2, coor_max_dim2)), patch_mask


from torch import nn
# FrEIA (https://github.com/VLL-HD/FrEIA/)
import FrEIA.framework as Ff
import FrEIA.modules as Fm


def subnet_fc(dims_in, dims_out):
    return nn.Sequential(nn.Linear(dims_in, 2*dims_in), nn.ReLU(), nn.Linear(2*dims_in, dims_out))


def flow_model(args, in_channels):
    coder = Ff.SequenceINN(in_channels)
    print('Normalizing Flow => Feature Dimension: ', in_channels)
    for k in range(args.coupling_layers):
        coder.append(Fm.AllInOneBlock, subnet_constructor=subnet_fc, affine_clamping=args.clamp_alpha,
            global_affine_type='SOFTPLUS', permute_soft=True)
    return coder


def conditional_flow_model(args, in_channels):
    coder = Ff.SequenceINN(in_channels)
    print('Conditional Normalizing Flow => Feature Dimension: ', in_channels)
    for k in range(args.coupling_layers):  # 8
        coder.append(Fm.AllInOneBlock, cond=0, cond_shape=(args.pos_embed_dim,), subnet_constructor=subnet_fc, affine_clamping=args.clamp_alpha,
            global_affine_type='SOFTPLUS', permute_soft=True)
    return coder